// name: Karplus.cpp

#include "Karplus.h"
#include <cstdlib>
#include <memory.h>
#include <assert.h>
#include <math.h>
#include <stdio.h>

#define MY_SRATE 44100

long g_delayMax = 200000;

// ctor

Karplus::Karplus()
{
   Karplus(60, 100); 
}


Karplus::Karplus( int midinum, int velocity)
{
    myMidinum = midinum;
    myVelocity = velocity;

    myAmplitude = (double) 0.5 * velocity / 128.0; 
    myAttenuation = .99;

    // read and write indices
    myReader = 0;
    myWriter = 0;
    mySampleIndex = 0;

    setDelayLen(); // sets myDelayLen based on midinum

    // (dynamically) allocate the delay buffer
    myDelay = new SAMPLE[g_delayMax];
    // zero out
    memset( myDelay, 0, g_delayMax*sizeof(SAMPLE) );
    assert( myDelayLen <= g_delayMax );
    // set indices
    myReader = myWriter - myDelayLen;
    // wrap it properly
    if( myReader < 0 ) myReader += g_delayMax;

    setDelay();     // populates myDelay 

}

// dtor
Karplus::~Karplus()
{
	delete [] myDelay;
}



// reads the next sample from myDelay
SAMPLE Karplus::getNextSample()
{
	SAMPLE value = myDelay[mySampleIndex];
	mySampleIndex++;
	if (mySampleIndex == g_delayMax ) {
	   mySampleIndex=0; 
   	   memset( myDelay, 0, g_delayMax*sizeof(SAMPLE) ); // zero out so that the note doesn't repeat every g_delayMax
	}
	return value; 

}

// converts midi number to frequency
double Karplus::mtof( int midi )
{
    if( midi < 20 ) return ( mtof(20));
    else if( midi > 110 ) return (mtof(110));
    else return ( pow(2,(midi-69)/12.0) * 440.0 );
}


// sets myDelayLen according to midinum
void Karplus::setDelayLen() 
{
	myDelayLen = MY_SRATE / mtof(myMidinum);
}


// populates myDelay vector
void Karplus::setDelay()
{
    srand ((unsigned)time(NULL));	// for white noise

    for (int i=0; i<g_delayMax; i++) {

	// load an array of white noise, of length g_delayLen;
        if (i < myDelayLen) myDelay[myWriter] = myAmplitude * ( ((float) rand()/RAND_MAX) *2 - 1 ); 
	else myDelay[myWriter] = 0;

	// moving-average filter
	myDelay[myWriter] += myAttenuation * .5 * (myDelay[myReader] + myDelay[(myReader-1) % g_delayMax]);

        // increment
        myWriter++; myReader++;

        // wrap indices
        myWriter %= g_delayMax;
        myReader %= g_delayMax;
    }
}


// for debugging, print every 2000 elements of myDelay vector
void Karplus::print()
{
	for (int i=0; i<g_delayMax; i=i+2000) {
		cerr << "myDelay[" << i << "] = " << myDelay[i] << endl;
	}
}





