/* ----------------------------------------------------------------------------
 * Jieun Oh
 *
 * Music 256 Final Project
 * December 2008
 *
 * midisort.cpp
 *
 * summary: visualizes and sonifies sorting algorithms
 * note: visualization code in OpenGL taken from sndpeek
 *
 * TO DO: 
 * - cleanup fn
 * - fix unpredictable bus errors :( 
 *
 * FOR DEMO:
 * - (set g_MIDIin = true) midi keyboard input
 * - (set g_MIDIin = false) comparison of 2 algorithms using manual population
 *
  -------------------------------------------------------------------------- */

#include <cstdlib>
#include <memory.h>
#include <iostream>
#include <math.h>
#include <time.h>
#include <vector>
#include <map>

#include "RtAudio.h"     // RtAudio
#include "RtMidi.h"      // RtMidi
#include "Karplus.h"

#include <GLUT/glut.h>   // OpenGL
#include <OpenGL/gl.h>   // OpenGL
#include <OpenGL/glu.h>  // OpenGL

using namespace std;


// ----------------------------------------------------------------------------
// function prototypes
// ----------------------------------------------------------------------------

void print();
void printAll();
void populateEvolution();
void printStage(int stage);
void playStage(int stage);
void setTitle();

void processCommandline (int argc, char ** argv);
void idleFunc( );
void displayFunc( );
void reshapeFunc( int width, int height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );
void initialize_graphics( );
void initialize_audio();
// cleanUp();


// ----------------------------------------------------------------------------
// global variables and #defines
// ----------------------------------------------------------------------------

enum sortAlgorithm {BUBBLESORT, SELECTIONSORT, INSERTIONSORT, SHELLSORT};
sortAlgorithm myAlgorithm = BUBBLESORT;

typedef vector<int> Stage;
vector <Stage> g_evolution;         // stores each stages required to sort
Stage g_notes;                      // stores midi notes to be sorted
char* g_title;                      // title of the sorting algorithm to use
int g_currStage = -1;               // for stepping through g_evolution when sonifying & visualizing
int g_goalStage = -1;				// stage to reach before pausing
int g_stepSize = 1;                 // how many stages to step through before stopping; controllable through keyboard input

unsigned int g_port = 1;            // port number for midi input. IF THIS DOESN'T WORK, CHANGE BACK TO 1 !!!
map <int, Karplus* >  karpli;       // stores instances of Karplus class
map <int, Karplus* >::iterator it ; // iterator for karpli

#define INC_VAL_MOUSE           0.05f // 1.0f
#define INC_VAL_KB              .025f

bool g_MIDIin = true;               // true means we're using MIDI input
int g_noteCount = 0;				// # of notes that have been added to g_notes, for debuging

vector<int> g_comparisons;			// stats: per stage
vector<int> g_swaps ;				// stats: per stage

// width and height of the window
GLsizei g_width = 800;
GLsizei g_height = 800; 
GLsizei g_last_width = g_width;
GLsizei g_last_height = g_height;

// fill mode
GLenum g_fillmode = GL_FILL;
// light 0 position
GLfloat g_light0_pos[4] = { 2.0f, 1.2f, 4.0f, 1.0f };
// light 1 parameters
GLfloat g_light1_ambient[] = { .2f, .2f, .2f, 1.0f };
GLfloat g_light1_diffuse[] = { 1.0f, 1.0f, 1.0f, 1.0f };
GLfloat g_light1_specular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
GLfloat g_light1_pos[4] = { -2.0f, 0.0f, -4.0f, 1.0f };
// modelview stuff
GLfloat g_angle_y = 0.0f;
GLfloat g_inc = 0.0f;
GLfloat g_eye_y = 0.0f;

// real-time audio
RtAudio g_audio;
GLboolean g_ready = TRUE; 
pthread_mutex_t g_mutex;
GLuint g_srate = 44100;

// the index associated with the waterfall
GLint g_wf = 0;

// rotation increments, set to defaults
GLfloat g_inc_val_mouse = INC_VAL_MOUSE;
GLfloat g_inc_val_kb = INC_VAL_KB;
GLfloat g_space = .15f; 


//-----------------------------------------------------------------------------
// Name: usage
// Desc: (for RtMidi) Warns and exits if incorrect port is specified
//-----------------------------------------------------------------------------
void usage( void ) {
	// Error function in case of incorrect command-line argument specifications.
	std::cout << "\nuseage: cmidiin <port>\n";
	std::cout << "    where port = the device to use (default = 0).\n\n";
	exit( 0 );
}

//-----------------------------------------------------------------------------
// Name: midiCallback
// Desc: (callback for RtMidi) listens for MIDI input messages and
// insert into map upon note-on; remove from map upon note-off
//-----------------------------------------------------------------------------
void midiCallback( double deltatime, std::vector< unsigned char > *message, void *userData )
{
	unsigned int nBytes = message->size();
	int midinote; 
	int velocity;

	// if a midi event happens, set up karplus appropriately
	if ( ((int)message->at(0))==144 ) {
				
		midinote = (int)message->at(1);
		velocity = (int)message->at(2);
		
		// note-off: remove from map!
		if ( velocity == 0) {
			Karplus * temp = karpli[midinote];
			karpli.erase(midinote);
			delete temp; 
		}
		
		// note-on: create Karplus instance (w proper midinum & vel), insert into map!
		else {
			Karplus * k = new Karplus(midinote, velocity);  // creating a Karplus object
			karpli[midinote] = k;                           // inserting to map
			g_notes.push_back(midinote);                    // for sorting!!!
			g_noteCount++;
			cout << g_noteCount << ": " << midinote << " has been played." << endl;
		}
	}
}

//-----------------------------------------------------------------------------
// Name: setupMidiInput()
// Desc: spawns midiCallback, where keyboard inputs are played (using Karplus) and saved into g_notes.
// alternatively, if midi device is *not* used (specificed by boolean g_MIDIin), manually populates g_notes
//-----------------------------------------------------------------------------
void setupMidiInput() 
{
	// enter here if we find midi device (default is set to 1; change g_port above as necessary!)
	if (g_MIDIin) {
		
		RtMidiIn *midiin = 0;
		// RtMidiIn constructor
		try { midiin = new RtMidiIn();}
		catch ( RtError &error ) {
			error.printMessage();
			exit( EXIT_FAILURE ); }
	
		// Error handling on available ports 
		// g_port should already be set (from main), if specified
		unsigned int nPorts = midiin->getPortCount();
		if ( g_port >= nPorts ) {
			delete midiin;
			std::cout << "Invalid port specifier!\n";
			usage();
		}
	
		try { midiin->openPort( g_port );}
		catch ( RtError &error ) {
			error.printMessage();
			goto cleanupMidi;}

		midiin->setCallback( &midiCallback );
	
		// Don't ignore sysex, timing, or active sensing messages.
		midiin->ignoreTypes( false, false, false );
	
		std::cout << "\nReading MIDI input ... press <enter> to quit.\n\n";
		char input;
		std::cin.get(input);
		
	cleanupMidi:
		delete midiin;
	}
	
	// not using midi device for getting notes; manual population of g_notes
	else {
		g_notes.push_back(70);
		g_notes.push_back(100);
		g_notes.push_back(60);
		g_notes.push_back(80);
		g_notes.push_back(90);
		g_notes.push_back(110);
		g_notes.push_back(50);
		g_notes.push_back(40);
		g_notes.push_back(30);
		g_notes.push_back(120);
		g_notes.push_back(10);
		g_notes.push_back(20);
	}
		
	// we now have all our notes to sort, so determine stages of the sorting process
	populateEvolution();
}



//-----------------------------------------------------------------------------
// Name: callKarplus
// Desc: callback for RtAudio
//-----------------------------------------------------------------------------
// call back for karplus
int callKarplus( void * out, void * in, unsigned int numFrames,
				double streamTime, RtAudioStreamStatus status, void * userData )
{
    // warn in case of over/underflow
    if ( status ) std::cout << "Stream over/underflow detected." << std::endl;
	
	// lock
	pthread_mutex_lock( &g_mutex );
	g_ready = FALSE;
	
    // cast pointer to (hopefully) properly typed pointer
    SAMPLE * buffo = (SAMPLE *)out;
	
    // loop over frames
    for( int i = 0; i < numFrames; i++ )
    {
		buffo[2*i] = buffo[2*i+1] = 0.0; 
			for ( it= karpli.begin(); it != karpli.end(); it++ ) {
				buffo[2*i] += ((*it).second)->getNextSample();
				buffo[2*i+1] = buffo[2*i];
			}
	}

	// set flag
    g_ready = TRUE;
    // unlock
    pthread_mutex_unlock( &g_mutex );	
	
    return 0;    
}


//-----------------------------------------------------------------------------
// Name: main( )
// Desc: entry point
//-----------------------------------------------------------------------------
int main (int argc, char ** argv)
{	
	processCommandline ( argc, argv ); 
		
	// initialize GLUT
	glutInit( &argc, argv );
	// double buffer, use rgb color, enable depth buffer
	glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
	// initialize the window size
	glutInitWindowSize( g_width, g_height );
	// set the window postion
	glutInitWindowPosition( 100, 100 );
	// create the window
	glutCreateWindow( "midisort" );
	
	// set the idle function - called when idle
	glutIdleFunc( idleFunc );
	// set the display function - called when redrawing
	glutDisplayFunc( displayFunc );
	// set the reshape function - called when client area changes
	glutReshapeFunc( reshapeFunc );
	// set the keyboard function - called on keyboard events
	glutKeyboardFunc( keyboardFunc );
	// set the mouse function - called on mouse stuff
	glutMouseFunc( mouseFunc );
	
	// do our own initialization
	initialize_graphics( );
    
	// intialize real-time audio; within this g_notes is populated
	initialize_audio();
	
	// uncomment to debug: print contents of g_evolution for debugging
	// printAll(); 
	
	glutMainLoop();
	
	return 0;
}

//-----------------------------------------------------------------------------
// Name: populateEvolution( )
// Desc: runs sort algorithm and populates g_evolution according to myAlgorithm
//-----------------------------------------------------------------------------
void populateEvolution()
{
	// BUBBLESORT, SELECTIONSORT, INSERTIONSORT, SHELLSORT

	if (myAlgorithm == BUBBLESORT) {
		assert (g_evolution.size()==0);
		assert (g_comparisons.size()==0);
		assert (g_swaps.size()==0);

		int i, j;
		for(i=0; i< g_notes.size(); i++) {
			int stageComparisons = 0;
			int stageSwaps = 0;  
			
			for(j=0; j<i; j++) {
				stageComparisons++;
				if(g_notes[i] < g_notes[j])
				{
					stageSwaps++;
					// swap
					int temp = g_notes[i];  
					g_notes[i] = g_notes[j];
					g_notes[j] = temp;
				}
			}
			// insert current g_notes as a stage in g_evolution
			g_evolution.push_back(g_notes);
			g_comparisons.push_back(stageComparisons);
			g_swaps.push_back(stageSwaps);
		}
	}
	
	else if (myAlgorithm == SELECTIONSORT) {
		assert (g_evolution.size()==0);
		assert (g_comparisons.size()==0);
		assert (g_swaps.size()==0);

		int i, j;
		g_evolution.push_back(g_notes);
		g_comparisons.push_back(0);
		g_swaps.push_back(0);
		
		for (i=0; i < g_notes.size()-1; i++) {
			int stageComparisons = 0;
			int stageSwaps = 0;  

			int potentialSmallest = i; 
			for (int j=i+1; j < g_notes.size(); j++) {
				stageComparisons++;
				if (g_notes[j] < g_notes[potentialSmallest]) {
					potentialSmallest = j;
				}
			}
			//swap
			stageSwaps++;
			int temp = g_notes[i];
			g_notes[i] = g_notes[potentialSmallest];
			g_notes[potentialSmallest] = temp;
			
			// insert current g_notes as a stage in g_evolution
			g_evolution.push_back(g_notes);
			g_comparisons.push_back(stageComparisons);
			g_swaps.push_back(stageSwaps);
		}
	}
	
	// add stats code for this!!!
	else if (myAlgorithm == INSERTIONSORT) {
		assert (g_evolution.size()==0);
		assert (g_comparisons.size()==0);
		assert (g_swaps.size()==0);
		
		int i, j, key;
		g_evolution.push_back(g_notes);
		g_comparisons.push_back(0);
		g_swaps.push_back(0);
		
		for (i=1; i<g_notes.size(); i++) {// starting with i=1 is intentional
			int stageComparisons = 0;
			int stageSwaps = 0;  
			
			key = g_notes[i];
			for (j=i-1; (j>=0) && (g_notes[j]>key); j--) {
				stageComparisons++;

				g_notes[j+1] = g_notes[j];
			}
			stageSwaps++;
			g_notes[j+1] = key;
			// insert current g_notes as a stage in g_evolution
			g_evolution.push_back(g_notes);
			g_comparisons.push_back(stageComparisons);
			g_swaps.push_back(stageSwaps);
		}
	}
	
	// add stats code for this!!!
	// Shell Sort, taken from http://en.wikipedia.org/wiki/Shell_sort 
	else {
		assert (myAlgorithm == SHELLSORT);
		assert (g_evolution.size()==0);
		assert (g_comparisons.size()==0);
		assert (g_swaps.size()==0);
		
		int i, j, increment, temp;
		g_evolution.push_back(g_notes);
		g_comparisons.push_back(0);
		g_swaps.push_back(0);
		
		increment = g_notes.size()/2;
		while (increment > 0) {
			for (i = increment; i<g_notes.size(); i++) {
				int stageComparisons = 0;
				int stageSwaps = 0;  
				
				j=i;
				temp = g_notes[i];
				while ((j>=increment) && (g_notes[j-increment]>temp)) {
					stageComparisons++;
					stageComparisons++;

					g_notes[j] = g_notes[j-increment];
					j = j-increment;
					stageSwaps++;
				}
				g_notes[j] = temp;
				g_evolution.push_back(g_notes);
				g_comparisons.push_back(stageComparisons);
				g_swaps.push_back(stageSwaps);
			}
			if (increment ==2) increment =1;
			else increment = (int) (increment / 2.2);
		}
	}
	
}

//-----------------------------------------------------------------------------
// Name: processCommandline(int argc, char ** argv)
// Desc: sets myAlgorithm to one of the enum SortAlgorithm types
//-----------------------------------------------------------------------------
void processCommandline (int argc, char ** argv)
{
	cerr << endl << "-----------------------------------------------------------------------------" << endl;
	cerr << "Welcome to midisort! " << endl
	     << "Type 'h' (when graphics window is active) for keyboard controls help menu." << endl;

	if (argc<2) {
		cerr << "   FYI: Using Bubble Sort as default." << endl
		<< "        To specify algorithm, type the following:" << endl
		<< "            ./midisort -bubble runs bubble sort" << endl
		<< "            ./midisort -selection runs selection esort" << endl
		<< "            ./midisort -insertion runs insertion sort" << endl
		<< "            ./midisort -shell runs shell sort" << endl;
	}
	
	else {
		string type = (string) argv[1]; 
		if (type == "-bubble") {
			myAlgorithm = BUBBLESORT;
			cerr << "You've specified to use Bubble Sort." << endl;
		}
		else if (type == "-selection") {
			myAlgorithm = SELECTIONSORT;
			cerr << "You've specified to use Selection Sort." << endl;
		}
		else if (type == "-insertion") {
			myAlgorithm = INSERTIONSORT;
			cerr << "You've specified to use Insertion Sort." << endl;
		}
		else if (type == "-shell") {
			myAlgorithm = SHELLSORT;
			cerr << "You've specified to use Shell Sort." << endl;
		}
		else {
			cerr << "   Your input for [algorithm] had no match, so we're using Bubble Sort." << endl
			<< "        To specify algorithm, type the following:" << endl
			<< "            ./midisort -bubble runs bubble sort" << endl
			<< "            ./midisort -selection runs selection esort" << endl
			<< "            ./midisort -insertion runs insertion sort" << endl
			<< "            ./midisort -shell runs shell sort" << endl;
			myAlgorithm = BUBBLESORT;
		}
	}
	cerr << "-----------------------------------------------------------------------------" << endl << endl;

}


//-----------------------------------------------------------------------------
// Name: initialize_audio( )
// Desc: set up audio capture and playback and initializes any application data
// Note: no user control over port number; this is as defined in g_port above
//-----------------------------------------------------------------------------
void initialize_audio()
{
	// test for number of devices
	if( g_audio.getDeviceCount() < 1 )
	{
		cerr << "no audio device found!\n" << endl;
		exit( 1 );
	}
	
	// parameters
	unsigned int bufferFrames = 512;
	RtAudio::StreamParameters iParams, oParams;
	iParams.deviceId = g_audio.getDefaultInputDevice();
	iParams.nChannels = 2; // stereo
	oParams.deviceId = g_audio.getDefaultOutputDevice();
	oParams.nChannels = 2;
    
	// open the RtAudio
	try {
		g_audio.openStream( &oParams, &iParams, RTAUDIO_FLOAT64, 44100, &bufferFrames, &callKarplus, &karpli );
	} catch( RtError & err ) {
		err.printMessage();
		exit(1);
	}
    
	// start the stream
	try {
		// start
		g_audio.startStream();
		setupMidiInput();				//within this spawns midiCallback fn
	} catch( RtError & err ) {
		// do stuff
		err.printMessage();
	}

}


//-----------------------------------------------------------------------------
// Name: initialize_graphics( )
// Desc: sets initial OpenGL states and initializes any application data
//-----------------------------------------------------------------------------
void initialize_graphics()
{
    // set the GL clear color - use when the color buffer is cleared
    glClearColor( 0.0f, 0.0f,0.0f, 1.0f );
    // set the shading model to 'smooth'
    glShadeModel( GL_SMOOTH );
    // enable depth
    glEnable( GL_DEPTH_TEST );
    // set the front faces of polygons
    glFrontFace( GL_CCW );
    // set fill mode
    glPolygonMode( GL_FRONT_AND_BACK, g_fillmode );
    // enable lighting
    glEnable( GL_LIGHTING );
    // enable lighting for front
    glLightModeli( GL_FRONT_AND_BACK, GL_TRUE );
    // material have diffuse and ambient lighting 
    glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
    // enable color
    glEnable( GL_COLOR_MATERIAL );
	
    // enable light 0
    glEnable( GL_LIGHT0 );
	
    // setup and enable light 1
    glLightfv( GL_LIGHT1, GL_AMBIENT, g_light1_ambient );
    glLightfv( GL_LIGHT1, GL_DIFFUSE, g_light1_diffuse );
    glLightfv( GL_LIGHT1, GL_SPECULAR, g_light1_specular );
    glEnable( GL_LIGHT1 );	
}


//-----------------------------------------------------------------------------
// Name: cleanUp( )
// Desc: cleans up dynamic memory for vArray
//-----------------------------------------------------------------------------
void cleanUp( )
{
    // TODO: !!!
}


//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( int w, int h )
{
    // save the new window size
    g_width = w; g_height = h;
    // map the view port to the client area
    glViewport( 0, 0, w, h );
    // set the matrix mode to project
    glMatrixMode( GL_PROJECTION );
    // load the identity matrix
    glLoadIdentity( );
    // create the viewing frustum
    gluPerspective( 45.0, (GLfloat) w / (GLfloat) h, 1.0, 300.0 );
    // set the matrix mode to modelview
    glMatrixMode( GL_MODELVIEW );
    // load the identity matrix
    glLoadIdentity( );
    // position the view point
    gluLookAt( 0.0f, 3.5f * sin( g_eye_y ), 3.5f * cos( g_eye_y ), 
			  0.0f, 0.0f, 0.0f, 
			  0.0f, ( cos( g_eye_y ) < 0 ? -1.0f : 1.0f ), 0.0f );
}

//-----------------------------------------------------------------------------
// Name: printStage(int stage)
// Desc: prints sequence of midi notes in the order stored in g_evolution[stage]
//-----------------------------------------------------------------------------
void printStage(int stage)
{
	assert (stage > -1 && stage < g_evolution.size());
	for (int pos = 0; pos < g_notes.size(); pos++)
		cerr << g_evolution[stage][pos] << " " ;
	cerr << endl;
}


//-----------------------------------------------------------------------------
// Name: playStage(int stage)
// Desc: plays sequence of midi notes in the order stored in g_evolution[stage]
//-----------------------------------------------------------------------------
void playStage(int stage)
{	
	if (!g_ready) return; // doesn't seem to affect :( but not fatal either way

	assert (stage > -1 && stage < g_evolution.size());
	int pos, midinote, velocity;
	for (pos = 0; pos < g_notes.size(); pos++) {
		
		midinote = g_evolution[stage][pos];
		velocity = 80;							// arbitrarily set
		
		// note-on: create Karplus instance (w proper midinum & vel), insert into map!
		Karplus * k = new Karplus(midinote, velocity);  // creating a Karplus object
		karpli[midinote] = k;                           // inserting to map
		
		usleep( 100000 ); // 100ms? = default note length
		
		// note-off: remove from map!
		Karplus * temp = karpli[midinote];
		karpli.erase(midinote);
		delete temp; 
	}
}


//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
	if (!g_ready) return; // ignore any keyboard input entered while music is playing
						  // doesn't seem to affect :( but not fatal either way
    switch( key )
    {
		// case '0' = replay the sound for current stage; case 1-9 steps forward
		case '0':
			fprintf( stderr, "replaying Stage %d: ", g_goalStage);
			printStage(g_goalStage);
			playStage(g_goalStage);
			break;
		case '1':
			g_goalStage = g_currStage +1;
			if (g_goalStage > g_evolution.size()-1) g_goalStage = g_evolution.size()-1;
			fprintf( stderr, "stepping 1 more step, to Stage %d: ", g_goalStage );
			printStage(g_goalStage);
			playStage(g_goalStage);
			break;
		case '2':
			g_goalStage = g_currStage +2;
			if (g_goalStage>g_evolution.size()-1) g_goalStage = g_evolution.size()-1;
			fprintf( stderr, "stepping 2 more steps, to Stage %d: ", g_goalStage);
			printStage(g_goalStage);
			playStage(g_goalStage);
			break;
		case '3':
			g_goalStage = g_currStage +3;
			if (g_goalStage>g_evolution.size()-1) g_goalStage = g_evolution.size()-1;
			fprintf( stderr, "stepping 3 more step, to Stage %d: ", g_goalStage );
			printStage(g_goalStage);
			playStage(g_goalStage);
			break;
		case '4':
			g_goalStage = g_currStage +4;
			if (g_goalStage>g_evolution.size()-1) g_goalStage = g_evolution.size()-1;
			fprintf( stderr, "stepping 4 more steps, to Stage %d: ", g_goalStage);
			printStage(g_goalStage);
			playStage(g_goalStage);
			break;
		case '5':
			g_goalStage = g_currStage +5;
			if (g_goalStage>g_evolution.size()-1) g_goalStage = g_evolution.size()-1;
			fprintf( stderr, "stepping 5 more step, to Stage %d: ", g_goalStage );
			printStage(g_goalStage);
			playStage(g_goalStage);
			break;
		case '6':
			g_goalStage = g_currStage +6;
			if (g_goalStage>g_evolution.size()-1) g_goalStage = g_evolution.size()-1;
			fprintf( stderr, "stepping 6 more steps, to Stage %d: ", g_goalStage);
			printStage(g_goalStage);
			playStage(g_goalStage);
			break;
		case '7':
			g_goalStage = g_currStage +7;
			if (g_goalStage>g_evolution.size()-1) g_goalStage = g_evolution.size()-1;
			fprintf( stderr, "stepping 7 more step, to Stage %d: ", g_goalStage );
			printStage(g_goalStage);
			playStage(g_goalStage);
			break;
		case '8':
			g_goalStage = g_currStage +8;
			if (g_goalStage>g_evolution.size()-1) g_goalStage = g_evolution.size()-1;
			fprintf( stderr, "stepping 8 more steps, to Stage %d: ", g_goalStage);
			printStage(g_goalStage);
			playStage(g_goalStage);
			break;
		case '9':
			g_goalStage = g_currStage +9;
			if (g_goalStage>g_evolution.size()-1) g_goalStage = g_evolution.size()-1;
			fprintf( stderr, "stepping 9 more steps, to Stage %d: ", g_goalStage);
			printStage(g_goalStage);
			playStage(g_goalStage);
			break;
			
		// case 'b' = start over from stage 0 (i.e. completely unsorted)
		case 'b':
			g_currStage = -1;
			g_goalStage = 0;
			fprintf( stderr, "starting over from Stage 0: ");
			printStage(g_goalStage);
			playStage(g_goalStage);
			break;
		// case 'f' = finishes the sorting (i.e. to completely sorted)
		case 'f':
			g_goalStage = g_evolution.size()-1;
			fprintf( stderr, "finishing until completely sorted, to Stage %d: ", g_goalStage);
			printStage(g_goalStage);
			playStage(g_goalStage);
			break;
			
		// case 'u' and 'i' for manipulating z-scale
		case 'u':
			g_space *= 1.02f;
			break;
		case 'i':
			g_space *= .98f;
			break;
			
		// cse '[' and ']' for manipulating eye level
		case '[':
			g_eye_y -= g_inc_val_kb;
			break;
		case ']':
			g_eye_y += g_inc_val_kb;
			break;
		
		// exits program
		case 'q':
			exit( 0 );
			break;
			
		case 'h':
			cerr << endl;
			cerr << "----------------------- Keyboard Controls Help Menu -------------------------" << endl;
			cerr << "Type number zero '0' to REPLAY the current stage." << endl;
			cerr << "Type a number from '1' to step to the next stage." << endl;
			cerr << "Type a number from '2' to '9' to specify numbers of stage to jump forward." << endl;
			cerr << "Type 'b' to start over from BEGINNING (i.e. original midi input)." << endl;
			cerr << "Type 'f' to FINISH sorting (i.e. completely sorted)." << endl;
			cerr << "Type 'u' to increase, 'i' to decrease depth scale." << endl;
			cerr << "Type '[' to increase, ']' to decrease eye level." << endl;
			cerr << "Type 'q' to quit program." << endl;
			cerr << "Type 'h' to see Keyboard Controls Help Menu." << endl;
			cerr << "-----------------------------------------------------------------------------" << endl << endl;
			break;
  	}
	
    // do a reshape since g_eye_y might have changed
    reshapeFunc( g_width, g_height );
    glutPostRedisplay( );
}

//-----------------------------------------------------------------------------
// Name: mouseFunc( )
// Desc: handles mouse stuff
//-----------------------------------------------------------------------------
void mouseFunc( int button, int state, int x, int y )
{
    if( button == GLUT_LEFT_BUTTON )
    {
        // rotate
        if( state == GLUT_DOWN )
            g_inc -= g_inc_val_mouse;
        else 
            g_inc += g_inc_val_mouse;
    }
    else if ( button == GLUT_RIGHT_BUTTON )
    {
        if( state == GLUT_DOWN ) 
            g_inc += g_inc_val_mouse;
        else 
            g_inc -= g_inc_val_mouse;
    }
    else 
        g_inc = 0.0f;
	
    glutPostRedisplay( );
}


//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
    // render the scene
    glutPostRedisplay( );
}


//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{	
	// clear the color and depth buffers
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	
	// save current matrix state
    glPushMatrix( );
	
	// rotate the sphere about y axis
	glRotatef( g_angle_y += g_inc, 0.0f, 1.0f, 0.0f );
	
	GLint k, i;
	
	// color the spectrum
	glColor3f( 0.4f, 1.0f, 0.4f );
	// set vertex normals
	glNormal3f( 0.0f, 1.0f, 0.0f );
	
	// reset drawing variables
	GLfloat x = -0.8f;
	GLfloat y = -0.8f;
	GLfloat inc = 2.0f / (g_notes.size()+1);
	
	glLineWidth( 2.0f );

	// save current matrix state
	glPushMatrix();
	// translate in world coordinate
	glTranslatef(x, y, 0.0);
	// scale it
	glScalef( inc , 1.0 , -g_space );
	
	
	// X-AXIS (at completely sorted)
	glColor3f (0.2f, 0.4f, 1.0f);
	glBegin(GL_LINE_STRIP);
	glVertex3f(0, 0, 0);
	glVertex3f(g_notes.size()-1, 0, 0);
	glEnd();
	// marks indices (red dots) x-axis units
	glColor3f(1.0, 0.0, 0.0);
	for (int xx=0; xx<g_notes.size(); xx++) {
		glBegin(GL_QUADS);
		glVertex3f(xx-.05, -.01, 0);
		glVertex3f(xx+.05, -.01, 0); 
		glVertex3f(xx+.05, .01, 0); 
		glVertex3f(xx-.05, .01, 0);
		glEnd();	
	}
	
	// Y-AXIS (at completely sorted)
	glColor3f (0.2f, 0.4f, 1.0f);
	glBegin(GL_LINE_STRIP);
	glVertex3f(0, 0, 0);
	glVertex3f(0, 1.2, 0);
	glEnd();
	// marks indices (red dots) y-axis units
	glColor3f(1.0, 0.0, 0.0);
	for (int yy=0; yy<13; yy++) {
		glBegin(GL_QUADS);
		glVertex3f(-.1, 0.1*(yy-.05), 0);
		glVertex3f(.1, 0.1*(yy-.05), 0); 
		glVertex3f(.1, 0.1*(yy+.05), 0); 
		glVertex3f(-.1, 0.1*(yy+.05), 0);
		glEnd();	
	}
	
	// Z-AXIS
	glColor3f (0.2f, 0.4f, 1.0f);
	glBegin(GL_LINE_STRIP);
	glVertex3f(0, 0, 0);
	glVertex3f(0, 0, g_evolution.size()-1);
	glEnd();
	// marks indices (red dots) z-axis units
	glColor3f(1.0, 0.0, 0.0);
	for (int zz=0; zz<g_evolution.size(); zz++) {
		glBegin(GL_QUADS);
		glVertex3f(0, -.01, zz-.05);
		glVertex3f(0, .01, zz-.05); 
		glVertex3f(0, .01, zz+.05); 
		glVertex3f(0, -.01, zz+.05);
		glEnd();	
	}
	
	// RENDER G_EVOLUTION
	for( k = 0; k< g_goalStage+1 ; k++ )
	{
		glColor3f( 1.0f, 0.8f, 0.2f );
		glBegin( GL_LINE_STRIP );
		for( GLint i = 0; i < g_notes.size(); i++) 
			glVertex3f( i, g_evolution[k][i]*.01, g_evolution.size()-k-1 /*g_notes.size()-k-1*/);
		glEnd();
	}
	g_currStage = g_goalStage;
	
	// display sorting title
	static char str[1024];   
	setTitle();
	sprintf( str, g_title);
	char *c = str;
	glPushMatrix();
	glLineWidth( 4.0f );
	glColor3f (0.4f, 1.0f, 0.4f);
	glTranslatef(g_notes.size()*0.2, 1.7, 0.0f);
	glScalef( g_notes.size()*0.0007, .0009f, .02f);
	for (c=str; *c != '\0'; c++) {
		glutStrokeCharacter(GLUT_STROKE_ROMAN, *c);
	}
	glPopMatrix();
	
	// label for x-axis
	static char str1[1024];   
	sprintf( str1, "Index in Array");
	char *c1 = str1;
	glPushMatrix();
	glLineWidth( 2.0f );
	glColor3f (0.2f, 0.4f, 1.0f);
	glTranslatef( -0.6, -0.88f, 0.0f);
	glScalef( 0.001, .0004f, .01f);
	for (c1=str1; *c1 != '\0'; c1++) {
		glutStrokeCharacter(GLUT_STROKE_ROMAN, *c1);
	}
	glPopMatrix();
	
	// label for y-axis
	static char str2[1024];   
	sprintf( str2, "Pitch (MIDI num)");
	char *c2 = str2;
	glPushMatrix();
	glColor3f (0.2f, 0.4f, 1.0f);
	glTranslatef(-0.88, -0.6, -0.0f);
	glScalef( 0.00045f, .00055f, .005f);
	glRotatef(90, 0, 0, 1);
	for (c2=str2; *c2 != '\0'; c2++) {
		glutStrokeCharacter(GLUT_STROKE_ROMAN, *c2);
	}
	glPopMatrix();
	
	// label for z-axis
	static char str3[1024];   
	sprintf( str3, "Sorting Stages");
	char *c3 = str3;
	glPushMatrix();
	glColor3f (0.2f, 0.4f, 1.0f);
	glTranslatef(x, -0.865 , -0.4f);
	glScalef( 0.003f, .0003f, .0009f);
	glRotatef(90, 0, 1, 0);
	for (c3=str3; *c3 != '\0'; c3++) {
		glutStrokeCharacter(GLUT_STROKE_ROMAN, *c3);
	}
	glPopMatrix();
	
	
	// Display Stage Stats
	if (g_currStage>-1) {
	static char str4[1024];   
	sprintf( str4, "Displaying through Stage %d out of %d", g_currStage, g_evolution.size()-1);
	char *c4 = str4;
	glPushMatrix();
	glLineWidth( 1.0f );
	glColor3f (0.4f, 1.0f, 0.4f);
	glTranslatef(-0.5, 0.7 , 0.0);
	glScalef( 0.0005f, .0003f, .001f);
	for (c4=str4; *c4 != '\0'; c4++) {
		glutStrokeCharacter(GLUT_STROKE_ROMAN, *c4);
	}
	glPopMatrix();
	}
	
	// Display Swaps Stats
	if (g_currStage>-1) {
	int swapsSoFar = 0;
	for (int stage = 0; stage < g_currStage+1; stage++) {
		swapsSoFar += g_swaps[stage];
	}
	static char str5[1024];   
	sprintf( str5, "%d swap(s) made in current stage. Total Swaps: %d", g_swaps[g_currStage], swapsSoFar);
	char *c5 = str5;
	glPushMatrix();
	glLineWidth( 1.0f );
	glColor3f (1.0f, 1.0f, 1.0f);
	glTranslatef(-0.9, -1.1 , 0.0);
	glScalef( 0.0005f, .0003f, .0009f);
	for (c5=str5; *c5 != '\0'; c5++) {
		glutStrokeCharacter(GLUT_STROKE_ROMAN, *c5);
	}
	glPopMatrix();
	}
	
	// Display Comparisons Stats
	if (g_currStage>-1) {
		int comparisonsSoFar = 0;
		for (int stage = 0; stage < g_currStage+1; stage++) {
			comparisonsSoFar += g_comparisons[stage];
		}
		static char str6[1024];   
		sprintf( str6, "%d comparison(s) made in current stage. Total Comparisons: %d", g_comparisons[g_currStage], comparisonsSoFar);
		char *c6 = str6;
		glPushMatrix();
		glLineWidth( 1.0f );
		glColor3f (1.0f, 1.0f, 1.0f);
		glTranslatef(-0.9, -1.2 , 0.0);
		glScalef( 0.0005f, .0003f, .0009f);
		for (c6=str6; *c6 != '\0'; c6++) {
			glutStrokeCharacter(GLUT_STROKE_ROMAN, *c6);
		}
		glPopMatrix();
	}	
	
	// restore matrix state
	glPopMatrix();
	
    // flush gl commands
    glFlush( );
    // swap the buffers
    glutSwapBuffers( );
}

//-----------------------------------------------------------------------------
// Name: print( )
// Desc: prints g_notes vector for debugging
//-----------------------------------------------------------------------------
void setTitle()
{
	if (myAlgorithm == SELECTIONSORT) g_title = "Selection Sort";
	else if (myAlgorithm == INSERTIONSORT) g_title = "Insertion Sort";
	else if (myAlgorithm == SHELLSORT) g_title = "Shell Sort";
	else g_title = "Bubble Sort"; // this is the default
}

//-----------------------------------------------------------------------------
// Name: print( )
// Desc: prints g_notes vector for debugging
//-----------------------------------------------------------------------------
void print()
{
	for (int i=0; i< g_notes.size(); i++) 
		cout << " " << g_notes[i];
	cout << endl;
}

//-----------------------------------------------------------------------------
// Name: printAll( )
// Desc: prints g_evolution vector for debugging
//-----------------------------------------------------------------------------
void printAll()
{
	for (int i=0; i< g_evolution.size(); i++) {
		Stage temp = g_evolution[i];
		for (int j=0; j<temp.size(); j++) {
			cout << " " << temp[j];
		}
		cout << endl;
	}
}
